/**
 * Klasse für die Bilderkennung.
 * Läd die zu überprüfenden Positionen und Farben aus seperaten Dateien 
 * und erhält ein Bild übergeben, welches es analysiert.
 */
class FigureDetector {
  
  protected ArrayList<PVector> boardRegions = new ArrayList<PVector>();
  protected ArrayList<PVector> borderRegions = new ArrayList<PVector>();
  protected color[] figureColors = new color[12];
  
  public FigureDetector() {
    loadBoardRegions();
    loadBorderRegions();
    loadFigureColors();
  }
  
  /**
   * Läd die Farbwerte der Figuren aus einer Datei.
   */
  public void loadFigureColors() {
    String[] lines = loadStrings("figure_colors.txt");
    for (int i = 0; i < min(lines.length, figureColors.length); i++)
      figureColors[i] = unhex(lines[i]);
  }
  
  /**
   * Läd die Positonen, die überprüft werden sollen auf dem Schachbrett.
   */
  public void loadBoardRegions() {
    loadRegions("board_regions.txt", boardRegions);
  }
  
  /**
   * Läd die Positonen, die überprüft werden sollen am Rand des Schachbrettes.
   */
  public void loadBorderRegions() {
    loadRegions("border_regions.txt", borderRegions);
  }
  
  /**
   * Läd Regionen aus einer angegebenen Datei in eine ArrayList<PVector> hinein.
   * Gerade Indizes der Liste sind also Startpositionen und ungerade Endpositionen von Rechtecken.
   * Die Datei hat das Format:
   * x_start, y_start, x_ende, y_ende\n
   * 
   * @param filePath Pfadangabe zur Datei
   * @param Liste, in welche geladen werden soll
   */
  protected void loadRegions(String filePath, ArrayList<PVector> regions) {
    String[] lines = loadStrings(filePath);
    for (int i = 0; i < lines.length; i++) {
      String[] p = split(lines[i], ',');
      if (p.length == 4) {
        regions.add(new PVector(float(p[0]), float(p[1])));
        regions.add(new PVector(float(p[2]), float(p[3])));
      }
    }
  }
  
    
  //Methoden die noch implementiert werden müssen:
  //public ? getFigures(PImage image) ? Analysiert ein gegebenes Bild auf Figuren und gibt die gefundenen Figuren zurück
  
  
  
  // Hier kommen ein paar Methoden, die vl für nützlich sind...
  
  /**
   * Gibt den pixel-array eines rechteckigen Ausschnitts eines Bildes zurück.
   * 
   * @param image das Ursprungsbild
   * @param x die x-Koordinate des Ausschnittes (obere Ecke links)
   * @param y die y-Koordinate des Ausschnittes (obere Ecke links)
   * @param w die Breite des Ausschnittes
   * @param h die Höhe des Ausschnittes
   * @return der pixel-array des Ausschnittes
   */
  public color[] getPixelSubset(PImage image, int x, int y, int w, int h) {
    PImage subset = image.get(x, y, w, h);
    subset.loadPixels();
    return subset.pixels;
  }
  
  /**
   * Gibt den Median bezüglich der Farben aus einem pixel-array zurück.
   * 
   * @param data der pixel-array
   * @return der Medien der Farben
   */
  public color getMedianColor(color[] data) {
    if (data.length <= 0)
      return 0;
    color[] sorted = sort(data);
    return sorted[sorted.length / 2];
  }
  
  /**
   * Gibt die Differenz zweier Farben zurück.
   * Der Algorithmus der Farbendifferenz kann verändert werden, damit er den Licht-/Bildverhältnissen eher entspricht.
   * Statt red(c1) - red(c2) wäre so z.B. auch brightness(c1) - brightness(c2) möglich und eine Gewichtung.
   * (Also z.B. Farbwert ist zu 100% wichtig, Sättigung zu 50%, Helligkeit zu 50% o.ä.; muss man rumexperimentieren.)
   * 
   * @param c1 erste Farbe
   * @param c2 zweite Farbe
   * @return die Differenz
   */
  public float colorDifference(color c1, color c2) {
    return sqrt(sq(red(c1) - red(c2)) + sq(green(c1) - green(c2)) + sq(blue(c1) - blue(c2)));
  }
 
}