import processing.serial.*;

/**
 * Klasse, dessen Instanzen über eine serielle Schnittstelle mit dem SChFiM-Interpreter (Arduino) kommunizieren.
 */
class SChFiMTransceiver {
  
  protected Serial serial = null;
  
  /**
   * Erstellt eine Instanz von SChFiMTransceiver.
   * Es wird die erste Schnittstelle aus der Liste der möglichen Schnittstellen gewählt und mit der Frequenz 9600Hz initialisiert.
   *
   * @param applet Referenz zum Hauptskript (Instanz von PApplet)
   */
  public SChFiMTransceiver(PApplet applet) {
    try {
      this.serial = new Serial(applet, Serial.list()[0], 9600);
    } catch (Exception ex) {
      this.disconnect();
    }
  }
  
  /**
   * Erstellt eine Instanz von SChFiMTransceiver.
   * 
   * @param serial Schnittstelle für die serielle Kommunikation
   */
  public SChFiMTransceiver(Serial serial) {
    this.serial = serial;
  }
  
  /**
   * Gibt zurück, ob eine Verbindung hergestellt wurde.
   */
  public boolean isConnected() {
    return this.serial != null;
  }
  
  /**
   * Bricht die aktuelle Verbindung ab und verbindet neu.
   *
   * @param serial Schnittstelle für die serielle Kommunikation
   */
  public void connect(Serial serial) {
    this.disconnect();
    this.serial = serial;
  }
  
  /**
   * Bricht die aktuelle Verbindung ab.
   */
  public void disconnect() {
    if (this.serial != null)
      this.serial.stop();
    this.serial = null;
  }
  
  /**
   * Gibt die Schnittstelle für die serielle Kommunikation zurück.
   * 
   * @return die Schnittstelle; null, falls keine Verbindung geschafft wurde
   */
  public Serial getSerial() {
    return this.serial;
  }
  
  /**
   * Sendet Daten über die serielle Schnittstelle.
   * Bricht die Verbindung ab, falls ein Fehler auftritt.
   * 
   * @param data die Daten, welche gesendet werden sollen
   * @return Wahrheitswert, ob das Senden erfolgreich war
   */
  public boolean transmit(String data) {
    if (this.serial == null)
      return false;
    print(this.serial.active());
    try {
      this.serial.write(data);
    } catch (Exception ex) {
      print(ex);
      this.disconnect();
      return false;
    }
    return true;
  }
  
  /**
   * Sendet einen Schachzug über die serielle Schnittstelle.
   * Bricht die Verbindung ab, falls ein Fehler auftritt.
   * 
   * @param movement der Schachzug, welcher gesendet werden soll
   * @return Wahrheitswert, ob das Senden erfolgreich war
   */
  public boolean transmit(ChessMovement movement) {
    return this.transmit(movement.toSChFiMCommandSequence());
  }
  
  /* Request Data, Set Data, Get Data
   * Buffered by Queue or simmilar
   */
  
  @Deprecated
  public boolean poll() {
    return false;
  }
  
  /**
   * Sucht im übergebenen String nach Daten und versucht diese zu interpretieren.
   * 
   * @param data String mit Daten
   */
  protected void interpretLines(String data) {
    
  }
  
}